// (c) Copyright 2013 Cloudera, Inc. All rights reserved.
define([
  'knockout',
  'underscore'
], function(ko, _) {

  // This is only used locally by the FilterModel::values property. If this
  // ends up being generally useful we can extract it to its own file.
  ko.extenders.deepArrayEquals = function(target) {
    return ko.computed({
      // Always return the target value.
      read: target,

      write: function(newValue) {
        var current = target();

        // Only set if the two values are not already equal.
        if (_.isArray(newValue) && !_.isEqual(newValue, current)) {
          target(newValue);
        }
      }
    });
  };

  var FilterModel = function(searchFilters) {
    var self = this;
    self.filter = ko.observable();
    self.compareType = ko.observable();
    // So... this is really stupid. The select2 control and the
    // selectedOptions binding don't play well together. When used together
    // they create an infinite loop of updates on initial construction of
    // the control. Apparently KO doesn't recognize that one empty array
    // is equal to another empty array; when initially constructed the
    // select2 control sets the <select> element to empty, which triggers
    // the observableArray to notify observers, which makes the ko.select2
    // binding trigger a change event on select2, which makes it set the
    // <select> element to empty... on and on it goes. I have used an
    // extender on this observableArray to only notify when a change
    // actually changes the array. It also double-checks that the value
    // assigned to the observableArray is an array.
    self.values = ko.observableArray().extend({ deepArrayEquals: null });
    // A filter is active if the user is currently editing it.
    self.active = ko.observable();

    // The following computed observables make templating easier.
    self.filterName = ko.computed({
      read: function() {
        var filter = self.filter();
        return filter ? filter.propertyName : null;
      },

      write: function(newValue) {
        if (!searchFilters) {
          return;
        }
        // Find the filter in the existing list of filters.
        var filter = _.find(searchFilters.filters(), function(filter) {
          return filter.propertyName === newValue;
        });
        self.filter(filter);
      }
    });

    self.compareTypeName = ko.computed({
      read: function() {
        var compareType = self.compareType();
        return compareType ? compareType.name : null;
      },

      write: function(newValue) {
        if (!self.filter()) {
          return;
        }
        var compareType = _.find(self.filter().compareTypes, function(compareType) {
          return compareType.name === newValue;
        });
        self.compareType(compareType);
      }
    });

    self.value = ko.computed({
      read: function() {
        return self.values().join(',');
      },
      write: function(newValue) {
        if (newValue) {
          var values = newValue.split(',');
          self.values(values);
        } else {
          self.values([]);
        }
      }
    });

    self.filter.subscribe(function(newFilter) {
      if (!newFilter || !newFilter.compareTypes ||
          newFilter.compareTypes.length !== 1) {
        return;
      }
      var compareType = newFilter.compareTypes[0];
      self.compareType(compareType);
    });

    // A FilterModel instance is complete if it has a filter, a compare type,
    // and at least one value in the values array.
    self.complete = ko.computed(function() {
      return self.filter() && self.compareType() &&
        self.values() && self.values().length;
    });
    
    self.activate = function(data, e) {
      if (!self.active() && !$(e.target).is('.btn')) {
        self.active(true);
      }
    };

    self.compareTo = function(to) {
      return to && self.filter() === to.filter()
          && self.compareType() === to.compareType()
          && _.isEqual(self.values().sort(), to.values().sort());
    };
  };

  return FilterModel;
});