// (c) Copyright 2013 Cloudera, Inc. All rights reserved.
define([
  'cloudera/search/FilterModel',
  'cloudera/Util',
  'knockout',
  'underscore'
], function(FilterModel, Util, ko, _) {
  // This function is bound to every filter object we get from the server.
  var has = function() {
    var self = this;
    // All of the attributes must be found to return true.
    return _.all(arguments, function(attr) {
      return self.hasOwnProperty(attr);
    });
  };

  // This function is bound to every filter object we get from the server.
  var hasnt = function() {
    var self = this;
    // None of the attributes must be found to return true.
    return !_.all(arguments, function(attr) {
      return self.hasOwnProperty(attr);
    });
  };

  // This function is bound to a filter object we get from the server if the
  // filter has an attribute 'valueUri'; this means that it is a
  // auto-complete filter and should query the server for values to help
  // the user.
  var executeQuery = function(query) {
    // Do nothing if the user hasn't entered anything yet.
    if (!query.term) {
      return;
    }
    var args = {
      q: query.term
    };
    $.getJSON(this.valueUri, args, function(data) {
      // The server is sending back a simple list of strings. We have
      // to decompose it for select2 in a particular way.
      var results = _.map(data, function(result) {
        return {
          id: result,
          text: result
        };
      });
      query.callback({
        results: results
      });
    });
  };

  // Manages a list of filters for use in searching.
  // The list of filters is retrieved from the server. The user chooses which
  // filters to apply at any given time.
  var SearchFilters = function(filterContainer, filterPath, staticFilters, exportResults) {
    this.filterContainer = filterContainer;
    this.filters = ko.observableArray();
    this.staticFilters = staticFilters || [];
    this.appliedFilters = ko.observableArray();
    this.filterPath = filterPath;
    this.activeFilter = ko.observable();
    this.loading = ko.observable(false);
    this.exportResults = exportResults;

    // These functions are used as Knockout callbacks and, thus, need a stable
    // reference to this.
    var self = this;
    self.addAnother = function() {
      self.deactivateAllFilters();
      self.appendFilter();
    };
    self.applyFilter = function(filter, compareType, value) {
      var filterModel = new FilterModel(self);
      if (filter) {
        filterModel.filter(filter);
        if (compareType) {
          if (_.isString(compareType)) {
            compareType = self.compareTypeFromName(filter, compareType);
            if (!compareType) {
              return false;
            }
          }
          filterModel.compareType(compareType);
          if (value) {
            if (!_.isArray(value)) {
              value = [value];
            }
            filterModel.values(value);
          }
        }
      }
      return this.appendFilter(filterModel);
    };
    self.removeFilter = function(removeMe) {
      // I'm not using _.filter here because I don't want to completely
      // erase the array. Knockout observable arrays take a bit of care.
      self.appliedFilters.remove(removeMe);
      // Are we removing the active filter? If so, deactivate it.
      if (removeMe && removeMe.compareTo(self.activeFilter())) {
        self.activeFilter(null);
      }
      $.publish('searchFilters.appliedFilters.changed');
      return false;
    };
    self.deactivateAllFilters = function() {
      var active = self.activeFilter();
      if (active) {
        active.active(false);
      }
    };

    ko.applyBindings(this, filterContainer);
  };

  // Bound to an instance of SearchFilters in the method appendFilter below.
  var onFilterModelActiveChanged = function(filterModel, newValue) {
    var currentlyActive = this.activeFilter();
    // Is a filter becoming active?
    if (newValue) {
      // Deactivate current filter if not same as filterModel.
      if (currentlyActive && currentlyActive !== filterModel) {
        currentlyActive.active(false);
      }
      // This is the new active filter.
      this.activeFilter(filterModel);
    } else {
      // Is the currently active the one that isn't active anymore?
      if (currentlyActive === filterModel) {
        if (!currentlyActive.complete()) {
          this.removeFilter(currentlyActive);
        }
        this.activeFilter(null);
      }
      // If this filter is no longer active but is complete, fire
      // the appliedfilters.change event.
      if (filterModel.complete()) {
        $.publish('searchFilters.appliedFilters.changed');
      }
    }
  };

  SearchFilters.prototype = {
    compareTypeFromName : function(filter, name) {
      return _.find(filter.compareTypes, function(compareType) {
        return compareType.name === name;
      });
    },
    loadFilters : function(callback) {
      $.getJSON(this.filterPath, _.bind(this.loadFiltersHandler, this))
        .complete(callback);
    },
    loadFiltersHandler: function(data) {
      var response = Util.filterError(data);
      this.clearFilters();
      this.registerAllFilters(response.data);
    },
    getAppliedFilters : function() {
      // Return a set of applied filters starting with a copy of any
      // static filters.
      var results = this.staticFilters.slice(0);
      _.each(this.appliedFilters(), function(filter) {
        // Make sure the filter is complete and that its compareType
        // has a name before returning it.
        if (filter.complete() && filter.compareType().name) {
          results.push({
            propertyName: filter.filter().propertyName,
            compareType: filter.compareType().name,
            values: filter.values()
          });
        }
      });
      return results;
    },
    appendFilter: function(filterModel) {
      // If there are any duplicates, stop here.
      var areEqual = function(filter) {
        return filter.compareTo(filterModel);
      };
      if (_.any(this.appliedFilters(), areEqual)) {
        return false;
      }
      // Subscribe to its active observable.
      var activeChanged = _.bind(onFilterModelActiveChanged, this, filterModel);
      filterModel.active.subscribe(activeChanged);
      this.appliedFilters.push(filterModel);
      // If it's not complete, then it is the new active filter.
      filterModel.active(!filterModel.complete());
      return true;
    },
    findFilter : function(name) {
      return _.find(this.filters(), function(filter) {
        return filter.propertyName === name;
      });
    },
    registerFilter : function(filter) {
      if (_.find(this.staticFilters, function(staticFilter) {
        return staticFilter.propertyName === filter.propertyName;
      })) {
        return;
      }
      // Assign some utility methods to the filter object for use in
      // the KO bindings in the template.
      filter.has = has;
      filter.hasnt = hasnt;
      if (filter.valueUri) {
        filter.executeQuery = _.bind(executeQuery, filter);
      }

      this.filters.push(filter);
    },
    registerAllFilters : function(filters) {
      _.each(filters, this.registerFilter, this);
    },
    clearFilters : function() {
      this.filters.splice(0);
    }
  };

  return SearchFilters;
});